#include "FLAME.h"

#define DD(i,j) buf_D[ (j)*ldim_D + (i) ]
#define UU(i,j) buf_U[ (j)*ldim_U + (i) ]
#define LL(i,j) buf_L[ (j)*ldim_L + (i) ]
#define pp(i) buf_p[ i ]

#define dabs(x) ( x < 0 ? -x : x )

int FLA_LU_SA_LAP_unb( FLA_Obj U, FLA_Obj D, FLA_Obj p, FLA_Obj L )
{
  int 
    m_U, m_D, *buf_p, i_one = 1, j, ipiv, m_U_min_j, m_U_min_j_min_1, 
    ldim_U, ldim_D, ldim_L;
  
  FLA_Datatype datatype;

  m_U = FLA_Obj_length( U );
  m_D = FLA_Obj_length( D );
  
  if ( m_D == 0 || m_U == 0 ) return FLA_SUCCESS;

  ldim_U = FLA_Obj_ldim( U );
  ldim_D = FLA_Obj_ldim( D );
  ldim_L = FLA_Obj_ldim( L );

  datatype = FLA_Obj_datatype( U );

  FLA_Copy( U, L );
  FLA_Triangularize( FLA_UPPER_TRIANGULAR, FLA_NONUNIT_DIAG, L );

  buf_p = ( int * ) FLA_Obj_buffer( p );

  if ( datatype == FLA_DOUBLE ){
    double 
      *buf_U, *buf_D, *buf_L,
      d_one = 1.0, d_minus_one = -1.0,
      d_inv_Ljj;


    buf_U = ( double * ) FLA_Obj_buffer( U );
    buf_D = ( double * ) FLA_Obj_buffer( D );
    buf_L = ( double * ) FLA_Obj_buffer( L );

    for ( j = 0; j<m_U; j++ ){
      ipiv = FLA_C2F( idamax )( &m_D, &DD( 0, j ), &i_one )-1;

      if ( dabs( LL( j, j ) ) < dabs( DD( ipiv, j ) ) ){
	FLA_C2F( dswap )( &m_U, &LL( j, 0 ), &ldim_L, &DD( ipiv, 0 ), &ldim_D ); 
	pp( j ) = ipiv + m_U - j;
      }	
      else pp( j ) = 0;

      d_inv_Ljj = d_one / LL( j, j );
      FLA_C2F( dscal )( &m_D, &d_inv_Ljj, &DD( 0, j ), &i_one ); 
      m_U_min_j_min_1 = m_U - j - 1;
      if ( m_U_min_j_min_1 > 0  )
	FLA_C2F( dger )( &m_D, &m_U_min_j_min_1, &d_minus_one, 
			 &DD( 0, j ), &i_one, &LL( j, j+1 ), &ldim_L,
			 &DD( 0, j+1 ), &ldim_D );

      m_U_min_j = m_U - j;
      if ( m_U_min_j > 0 ) 
	FLA_C2F( dcopy )( &m_U_min_j, &LL( j, j ), &ldim_L, &UU( j, j ), &ldim_U );

    }		 
  }

  return FLA_SUCCESS;
}
